/*
 * iflib_arith.c - Arithmetic conversions
 *
 * Written 2001 by Werner Almesberger
 * Copyright 2001 EPFL-ICA, Network Robots
 */


#include <assert.h>

#include "config.h"
#include "registry.h"
#include "data.h"
#include "op.h"
#include "iflib.h"


/*
 * General naming conventions in this file:
 *
 * X = general expression
 * C = constant
 * rel = relational operator
 */

/*
 * @@@ several of these optimizations should be either done in op.c, or the
 * optimizations in op.c should move here.
 */


/* ----- Helper functions -------------------------------------------------- */


static int log_2(uint32_t v)
{
    int ld = 0;

    if (!v) return -1;
    while (!(v & 1)) {
	ld++;
	v >>= 1;
    }
    return v == 1 ? ld : -1;
}


/* ----- *, /, and % ------------------------------------------------------- */


/*
 * Convert:
 *
 *  0 * X  ->  0
 *  X * 0  ->  0
 *  1 * X  ->  X
 *  X * 1  ->  X
 *  (2^n) * X  ->  X << n
 *  X * (2^n)  ->  X << n
 *
 *  X / 1      ->  X
 *  0 / X      ->  0
 *  X / X      ->  1
 *  X / (2^n)  ->  X >> n
 *
 *  X % 1      ->  0
 *  X % (2^n)  ->  X & (2^n)-1
 *  X % X      ->  0
 */


static int mul_opt(DATA *d,DATA x,DATA *c,uint8_t value)
{
    int ld;

    if (!value) {
	data_destroy(*d);
	*d = data_unum(0);
	return 1;
    }
    ld = log_2(value);
    if (ld == -1) return 0;
    data_destroy(*c);
    data_destroy_1(*d);
    if (!ld) {
	*d = x;
	return 1;
    }
    *d = op_binary(&op_shift_left,x,data_unum(ld));
    return 1;
}


static int div_l_opt(DATA *d,DATA x,DATA *c,uint8_t value)
{
    if (value) return 0;
    data_destroy(*d);
    *d = data_unum(0);
    return 1;
}


static int div_r_opt(DATA *d,DATA x,DATA *c,uint8_t value)
{
    int ld;

    ld = log_2(value);
    if (ld == -1) return 0;
    data_destroy(*c);
    data_destroy_1(*d);
    if (!ld) {
	*d = x;
	return 1;
    }
    *d = op_binary(&op_shift_right,x,data_unum(ld));
    return 1;
}


static int mod_opt(DATA *d,DATA x,DATA *c,uint8_t value)
{
    int ld;

    if (value == 1) {
	data_destroy(*d);
	*d = data_unum(0);
	return 1;
    }
    ld = log_2(value);
    if (ld == -1) return 0;
    d->op->dsc = &op_and;
    c->u.unum--;
    return 1;
}


static void mul_div_mod(DATA *d)
{
    if (!d->op) return;
    mul_div_mod(&d->op->a);
    mul_div_mod(&d->op->b);
    mul_div_mod(&d->op->c);
    if (d->op->dsc == &op_mult) {
	if (!d->op->a.op && d->op->a.type == dt_unum) {
	    if (mul_opt(d,d->op->b,&d->op->a,d->op->a.u.unum)) mul_div_mod(d);
	}
	else if (!d->op->b.op && d->op->b.type == dt_unum) {
	    if (mul_opt(d,d->op->a,&d->op->b,d->op->b.u.unum)) mul_div_mod(d);
	}
    }
    else if (d->op->dsc == &op_div) {
	if (!d->op->a.op && d->op->a.type == dt_unum) {
	    if (div_l_opt(d,d->op->b,&d->op->a,d->op->a.u.unum))
		mul_div_mod(d);
	}
	else if (!d->op->b.op && d->op->b.type == dt_unum) {
	    if (div_r_opt(d,d->op->a,&d->op->b,d->op->b.u.unum))
		mul_div_mod(d);
	}
	else if (expr_equal(d->op->a,d->op->b)) {
		    data_destroy(*d);
		    *d = data_unum(1);
	    }
    }
    else if (d->op->dsc == &op_mod) {
	if (!d->op->b.op && d->op->b.type == dt_unum) {
	    if (mod_opt(d,d->op->a,&d->op->b,d->op->b.u.unum)) mul_div_mod(d);
	}
	else if (expr_equal(d->op->a,d->op->b)) {
		data_destroy(*d);
		*d = data_unum(0);
	    }
    }
}


/* ----- Relational operators, common functions ---------------------------- */


static int is_relop(const OP_DSC *dsc)
{
    return dsc == &op_eq || dsc == &op_ne || dsc == &op_gt || dsc == &op_ge ||
      dsc == &op_lt || dsc == &op_le;
}


/* ----- Relational operators (move constants) ----------------------------- */


/*
 * Convert:
 *
 * C1 rel C2  ->  0  if !(C1 rel C2)
 * C1 rel C2  ->  1  if C1 rel C2
 *
 * C rel X  ->  X inv rel C  (exchange < and >, <= and >=)
 *
 * X+C1 rel C2  ->  X rel C1-C2
 * C1+X rel C2  ->  X rel C1-C2
 *
 * X-C1 rel C2  ->  X rel C1+C2
 * C1-X rel C2  ->  X rel C1-C2
 */


static void rel_plus(DATA *d,DATA *a,DATA *b)
{
    DATA *a_c,x;

    if (!a->op->a.op && a->op->a.type == dt_unum) {
	a_c = &a->op->a;
	x = a->op->b;
    }
    else if (!a->op->b.op && a->op->b.type == dt_unum) {
	x = a->op->a;
	a_c = &a->op->b;
    }
    else return;
    b->u.unum -= a_c->u.unum;
    data_destroy_1(*a_c);
    data_destroy_1(*a);
    *a = x;
}


static void rel_minus(DATA *d,DATA *a,DATA *b)
{
    DATA *a_c,x;

    if (!a->op->a.op && a->op->a.type == dt_unum) {
	a_c = &a->op->a;
	x = a->op->b;
	b->u.unum = a_c->u.unum-b->u.unum;
    }
    else if (!a->op->b.op && a->op->b.type == dt_unum) {
	x = a->op->a;
	a_c = &a->op->b;
	b->u.unum += a_c->u.unum;
    }
    else return;
    data_destroy_1(*a_c);
    data_destroy_1(*a);
    *a = x;
}


static void rel_math(DATA *d)
{
    if (!d->op) return;
    if (is_relop(d->op->dsc)) {
	if (!d->op->a.op && !d->op->b.op && d->op->a.type == dt_unum &&
	  d->op->b.type == dt_unum) {
	    DATA tmp = op_binary(d->op->dsc,d->op->a,d->op->b);

	    data_destroy(*d);
	    *d = tmp;
	    return;
	}
	if (!d->op->a.op && d->op->a.type == dt_unum && d->op->b.op) {
	    DATA tmp;

	    tmp = d->op->a;
	    d->op->a = d->op->b;
	    d->op->b = tmp;
	    if (d->op->dsc == &op_gt) d->op->dsc = &op_lt;
	    else if (d->op->dsc == &op_lt) d->op->dsc = &op_gt;
	    else if (d->op->dsc == &op_ge) d->op->dsc = &op_le;
	    else if (d->op->dsc == &op_le) d->op->dsc = &op_ge;
	}
	if (d->op->a.op && d->op->a.op->dsc == &op_plus && !d->op->b.op &&
	  d->op->b.type == dt_unum) {
	    rel_plus(d,&d->op->a,&d->op->b);
	    return;
	}
	if (d->op->a.op && d->op->a.op->dsc == &op_minus && !d->op->b.op &&
	  d->op->b.type == dt_unum) {
	    rel_minus(d,&d->op->a,&d->op->b);
	    return;
	}
    }
    rel_math(&d->op->a);
    rel_math(&d->op->b);
    rel_math(&d->op->c);
}


/* ----- Relational operators (canonical mask and shift) ------------------- */


/*
 * Convert:
 *
 * X rel C  ->  X & Cg rel C  where Cg is the combined mask of X
 *
 * At each step:
 *
 * X & Cm rel C  ->  X rel C  and  Cg = Cg & Cm
 * X << Cs rel C  ->  X rel C >> Cs  and  Cg = Cg >> Cs
 * X >> Cs rel C  ->  X rel C << Cs  and  Cg = Cg << Cs
 *
 * If C << Cs exceeds 32 bits,  X rel C  becomes  0 rel 1
 * If Cg < C,  X rel C  becomes  0 rel 1
 * If Cg becomes 0,  X rel C  becomes  0 rel C
 */


static void rel_access(DATA *d)
{
    DATA *x,old;
    uint32_t c,mask;
    int c_shift;

    if (!d->op) return;
    if (d->op->dsc == &op_logical_or || d->op->dsc == &op_logical_and ||
      d->op->dsc == &op_logical_not || d->op->dsc == &op_offset ||
      d->op->dsc == &op_precond) {
	rel_access(&d->op->a);
	rel_access(&d->op->b);
	return;
    }
    if (d->op->dsc == &op_conform || d->op->dsc == &op_count) return;
    if (!is_relop(d->op->dsc)) *d = op_binary(&op_ne,*d,data_unum(0));
    old = *d;
    if (d->op->b.op || d->op->b.type != dt_unum)
	dump_failed(*d,"\"rel const\" expected");
    c = d->op->b.u.unum;
    c_shift = 0;
    x = &d->op->a;
    mask = ~0;
    while (x->op) {
	uint32_t shift;

	if (x->op->dsc == &op_and) {
	    if (!x->op->a.op && x->op->a.type == dt_unum) {
		mask &= x->op->a.u.unum;
		x = &x->op->b;
		continue;
	    }
	    if (!x->op->b.op && x->op->b.type == dt_unum) {
		mask &= x->op->b.u.unum;
		x = &x->op->a;
		continue;
	    }
	    break;
	}
	if (x->op->b.op || x->op->b.type != dt_unum) break;
	shift = x->op->b.u.unum;
	if (x->op->dsc == &op_shift_left) {
	    mask = shift < 32 ? mask >> shift : 0;
	    c_shift -= shift;
	    if (!mask) break;
	    x = &x->op->a;
	    continue;
	}
	if (x->op->dsc == &op_shift_right) {
	    mask = shift < 32 ? mask << shift : 0;
	    c_shift += shift;
	    if (!mask) break;
	    x = &x->op->a;
	    continue;
	}
	break;
    }
    if (c_shift > 0) {
	/* overflow */
	if ((c_shift >= 32 && c) || c >> (32-c_shift)) {
	    *d = op_binary(d->op->dsc,data_unum(0),data_unum(1));
	    data_destroy(old);
	    return;
	}
	c <<= c_shift;
    }
    if (c_shift < 0) {
	uint32_t lost = c & (c_shift > -32 ? (1 << -c_shift)-1 : ~0);

	c = c_shift > -32 ? c >> -c_shift : 0;
	/* underflow */
	if (lost) {
	    /* X == C+eps  ->  0 */
	    if (d->op->dsc == &op_eq) {
		*d = data_unum(0);
		data_destroy(old);
		return;
	    }

	    /* X != C+eps  ->  1 */
	    if (d->op->dsc == &op_ne) {
		*d = data_unum(1);
		data_destroy(old);
		return;
	    }
	    /* X < C+eps  ->  X <= C */
	    if (d->op->dsc == &op_lt) d->op->dsc = &op_le;

	    /* X > C+eps  ->  X > C+1 */
	    if (d->op->dsc == &op_gt) c++;

	    /* X >= C+eps  ->  X > C */
	    if (d->op->dsc == &op_ge) d->op->dsc = &op_gt;

	    /* X <= C+eps  ->  X <= C, nothing to do */
	}
    }
    if (x->op && x->op->dsc == &op_access && x->op->c.u.unum &&
      x->op->c.u.unum < 32)
	mask &= (1 << x->op->c.u.unum)-1;
    if (mask < c) *d = op_binary(d->op->dsc,data_unum(0),data_unum(1));
    else if (!(mask | c)) *d = op_binary(d->op->dsc,data_unum(0),data_unum(0));
    else if ((c & ~mask) && (d->op->dsc != &op_lt && d->op->dsc != &op_gt))
	*d = data_unum(d->op->dsc == &op_ne);
    else *d = op_binary(d->op->dsc,
	  op_binary(&op_and,data_clone(*x),data_unum(mask)),
	  data_unum(c));
    data_destroy(old);
}


/* ----- Relational operators (turn them into equality) -------------------- */


/*
 * Convert:
 *
 * X < 2^n  ->  X & ~(2^n-1) == 0
 * X <= 2^n-1  ->  X & ~(2^n-1) == 0
 * X > 2^n-1  ->  X & ~(2^n-1) != 0
 * X >= 2^n  ->  X & ~(2^n-1) != 0
 */


static void rel_to_eq(DATA *d)
{
    if (!d->op) return;
    rel_to_eq(&d->op->a);
    rel_to_eq(&d->op->b);
    rel_to_eq(&d->op->c);
    if (d->op->a.op && !d->op->b.op && d->op->b.type == dt_unum) {
	uint32_t mask;
	int ld;

	if (d->op->dsc == &op_lt) {
	    ld = log_2(d->op->b.u.unum);
	    if (ld == -1) return;
	    mask = d->op->b.u.unum-1;
	    d->op->dsc = &op_eq;
	}
	else if (d->op->dsc == &op_le) {
	    ld = log_2(d->op->b.u.unum+1);
	    if (ld == -1 && (d->op->b.u.unum+1)) return; /* roll-over */
	    mask = d->op->b.u.unum;
	    d->op->dsc = &op_eq;
	}
	else if (d->op->dsc == &op_gt) {
	    ld = log_2(d->op->b.u.unum+1);
	    if (ld == -1 && (d->op->b.u.unum+1)) return; /* roll-over */
	    mask = d->op->b.u.unum;
	    d->op->dsc = &op_ne;
	}
	else if (d->op->dsc == &op_ge) {
	    ld = log_2(d->op->b.u.unum);
	    if (ld == -1) return;
	    mask = d->op->b.u.unum-1;
	    d->op->dsc = &op_ne;
	}
	else return;
	d->op->a = op_binary(&op_and,d->op->a,data_unum(~mask));
	d->op->b.u.unum = 0;
    }
}


/* ------ Eliminate expressions involving |, ^ ----------------------------- */


/*
 * Convert:
 *
 * X == X	 ->  1
 * X ^ C1 == C2  ->  X == C1 ^ C2
 * X | C1 == C2  ->  0				if (C1 & C2) != C1
 * X | C1 == C2  ->  X & ~C1 == (C2 & ~C1)	else
 *
 * ... also for C1 op X
 */


static void rel_eq_or(DATA *d)
{
    DATA x,c;

    if (!d->op) return;
    rel_eq_or(&d->op->a);
    rel_eq_or(&d->op->b);
    rel_eq_or(&d->op->c);
    if (d->op->dsc == &op_eq && expr_equal(d->op->a,d->op->b)) {
	data_destroy(*d);
	*d = data_unum(1);
	return;
    }
    if (d->op->dsc != &op_eq || d->op->b.op || d->op->b.type != dt_unum)
	return;
    if (!d->op->a.op || (d->op->a.op->dsc != &op_or &&
      d->op->a.op->dsc != &op_xor)) return;
    if (!d->op->a.op->a.op && d->op->a.op->a.type == dt_unum) {
	c = d->op->a.op->a;
	x = d->op->a.op->b;
    }
    else if (!d->op->a.op->b.op && d->op->a.op->b.type == dt_unum) {
	x = d->op->a.op->a;
	c = d->op->a.op->b;
    }
    else return;
    if (d->op->a.op->dsc == &op_xor) {
	d->op->b.u.unum ^= c.u.unum;
	data_destroy_1(d->op->a);
	data_destroy_1(c);
	d->op->a = x;
	return;
    }
    if ((c.u.unum & d->op->b.u.unum) != c.u.unum) {
	data_destroy(*d);
	*d = data_unum(0);
	return;
    }
    d->op->a.op->dsc = &op_and;
    d->op->a.op->a = x;
    d->op->a.op->b = data_unum(~c.u.unum);
    d->op->b.u.unum &= ~c.u.unum;
}


/* ----- General relational operators -------------------------------------- */


/*
 * The basic operation is  X <= C
 * Then, build_le can also build the reverse, i.e. X > C
 *
 * The following shortcuts are recognized:
 *
 * X <= C: 1 if C == 0xfffffff
 * X > C:  0 if C == 0xfffffff
 *
 * The two remaining relational operators are implemented as follows:
 *
 * X < C: 0	    if C == 0
 *        X <= C-1  if C != 0
 *
 * X >= C: 1	    if C == 0
 *         X > C-1  if C != 0
 *
 * Note that we always assume 32 bits, so later processing stages may
 * complain about "impossible" values.
 *
 * Also, we may end up with long redundant chains like  a || ... || x || 1
 * or  a && ... && x && 0
 */


static DATA build_le_bit(DATA acc,uint32_t c,int invert)
{
    DATA d,*next = &d;
    int i;

    for (i = 31; i >= 0; i--) {
	const OP_DSC *op;

	op = ((c >> i) & 1) == invert ? &op_logical_and : &op_logical_or;
	/*
	 * if invert:
	 *   next = (acc & (1 << i)) == (1 << i) op 0
	 * else:
	 *   next = (acc & (1 << i)) == 0 op 1
	 */
	*next = op_binary(op,
	  op_binary(&op_eq,
	    op_binary(&op_and,
	      data_clone(acc),
	      data_unum(1 << i)),
	    data_unum(invert ? 1 << i : 0)),
	  data_unum(1-invert));
	next = &next->op->b;
    }
    return d;
}


static void build_rel_bit(DATA *d,int invert,int const_res,uint32_t const_if,
  int decrement)
{
    uint32_t c;

    if (d->op->b.op || d->op->b.type != dt_unum) return; /* bad luck */
    c = d->op->b.u.unum;
    if (c == const_if) {
	data_destroy(*d);
	*d = data_unum(const_res);
    }
    else {
	DATA new;

	new = build_le_bit(d->op->a,decrement ? c-1 : c,invert);
	data_destroy(*d);
	*d = new;
    }
}


static void rel_general_bit(DATA *d)
{
    if (!d->op) return;
    if (d->op->dsc == &op_le) build_rel_bit(d,0,1,~0UL,0);
    else if (d->op->dsc == &op_gt) build_rel_bit(d,1,0,~0UL,0);
    else if (d->op->dsc == &op_lt) build_rel_bit(d,0,0,0,1);
    else if (d->op->dsc == &op_ge) build_rel_bit(d,1,1,0,1);
    if (d->op) {
	rel_general_bit(&d->op->a);
	rel_general_bit(&d->op->b);
	rel_general_bit(&d->op->c);
    }
}


/* ----- General relational operators, prefix version ---------------------- */


/*
 * Like rel_*_bit, but generates prefix matches, yielding a smaller number
 * of match operations, and a simpler expression structure, but each match
 * involves more bits.
 */


static DATA build_lt_prefix(DATA acc,uint32_t c,int invert)
{
    DATA d = data_none(),*next = &d;
    int i;

    for (i = 0; i < 32; i++) {
	if (((c >> i) & 1) == invert) continue;
	if (d.type != dt_none) {
	    *next = op_binary(&op_logical_or,*next,data_unum(0));
	    next = &next->op->b;
	}
	*next = op_binary(&op_eq,
	  op_binary(&op_and,
	    data_clone(acc),
	    data_unum(0xffffffff << i)),
	  data_unum((c & (0xffffffff << i))^(1 << i)));
    }
    assert(d.type != dt_none);
    return d;
}


static void build_rel_prefix(DATA *d,int invert,int const_res,uint32_t const_if,
  int xcrement)
{
    /*
     * if xcrement != 0:
     *   INcrement if !invert
     *   DEcrement if invert
     */
    uint32_t c;

    if (d->op->b.op || d->op->b.type != dt_unum) return; /* bad luck */
    c = d->op->b.u.unum;
    if (c == const_if) {
	data_destroy(*d);
	*d = data_unum(const_res);
    }
    else {
	DATA new;

	new = build_lt_prefix(d->op->a,xcrement ? c+(1-2*invert) : c,invert);
	data_destroy(*d);
	*d = new;
    }
}


static void rel_general_prefix(DATA *d)
{
    if (!d->op) return;
    if (d->op->dsc == &op_le) build_rel_prefix(d,0,1,~0UL,1);
    else if (d->op->dsc == &op_gt) build_rel_prefix(d,1,0,~0UL,0);
    else if (d->op->dsc == &op_lt) build_rel_prefix(d,0,0,0,0);
    else if (d->op->dsc == &op_ge) build_rel_prefix(d,1,1,0,1);
    if (d->op) {
	rel_general_prefix(&d->op->a);
	rel_general_prefix(&d->op->b);
	rel_general_prefix(&d->op->c);
    }
}


/* ----- Turn != into multiple == (bit or prefix) -------------------------- */


static DATA build_ne_bit(DATA acc,uint32_t c)
{
    DATA d = data_none(),*next = &d;
    int i;

    for (i = 0; i < 32; i++) {
	if (d.type != dt_none) {
	    *next = op_binary(&op_logical_or,*next,data_unum(0));
	    next = &next->op->b;
	}
	*next = op_binary(&op_eq,
	  op_binary(&op_and,
	    data_clone(acc),
	    data_unum(1 << i)),
	  data_unum(~c & (1 << i)));
    }
    return d;
}


static DATA build_ne_prefix(DATA acc,uint32_t c)
{
    DATA d = data_none(),*next = &d;
    int i;

    for (i = 0; i < 32; i++) {
	if (d.type != dt_none) {
	    *next = op_binary(&op_logical_or,*next,data_unum(0));
	    next = &next->op->b;
	}
	*next = op_binary(&op_eq,
	  op_binary(&op_and,
	    data_clone(acc),
	    data_unum(0xffffffff << i)),
	  data_unum((c & (0xffffffff << i))^(1 << i)));
    }
    return d;
}


static void build_ne(DATA *d)
{
    DATA new;
    uint32_t c;

    if (d->op->b.op || d->op->b.type != dt_unum) return; /* bad luck */
    c = d->op->b.u.unum;
    if (registry_probe(&optimization_switches,"prefix"))
	new = build_ne_prefix(d->op->a,c);
    else new = build_ne_bit(d->op->a,c);
    data_destroy(*d);
    *d = new;
}


static void ne_to_eq(DATA *d)
{
    if (!d->op) return;
    if (d->op->dsc == &op_ne) build_ne(d);
    if (d->op) {
	ne_to_eq(&d->op->a);
	ne_to_eq(&d->op->b);
	ne_to_eq(&d->op->c);
    }
}



/* ------------------------------------------------------------------------- */


static void rel_general(DATA *d)
{
    if (registry_probe(&optimization_switches,"prefix")) rel_general_prefix(d);
    else rel_general_bit(d);
}


void iflib_arith(DATA *d)
{
    mul_div_mod(d);
    debug_expr("AFTER mul_div_mod",*d);
    rel_math(d);
    debug_expr("AFTER rel_math",*d);
    rel_eq_or(d);
    debug_expr("AFTER rel_eq_or",*d);
    rel_access(d);
    debug_expr("AFTER rel_access",*d);
    rel_to_eq(d);
    debug_expr("AFTER rel_to_eq",*d);
    rel_general(d);
    debug_expr("AFTER rel_general",*d);
    if (registry_probe(&optimization_switches,"ne")) {
	ne_to_eq(d);
	debug_expr("AFTER ne_to_eq",*d);
    }
}
